@@ -1,5 +1,5 @@
 package Net::Async::Webservice::DHL::Address;
-$Net::Async::Webservice::DHL::Address::VERSION = '0.01_6';
+$Net::Async::Webservice::DHL::Address::VERSION = '0.001';
 {
   $Net::Async::Webservice::DHL::Address::DIST = 'Net-Async-Webservice-DHL';
 }
@@ -11,6 +11,15 @@ use Net::Async::Webservice::DHL::Types ':types';
 # ABSTRACT: an address for DHL
 
 
+for my $l (1..3) {
+    has "line$l" => (
+        is => 'ro',
+        isa => Str,
+        required => 0,
+    );
+}
+
+
 has city => (
     is => 'ro',
     isa => Str,
@@ -18,6 +27,13 @@ has city => (
 );
 
 
+has division => (
+    is => 'ro',
+    isa => Str,
+    required => 0,
+);
+
+
 has postal_code => (
     is => 'ro',
     isa => Str,
@@ -27,28 +43,55 @@ has postal_code => (
 
 has country_code => (
     is => 'ro',
-    isa => Str,
+    isa => CountryCode,
     required => 1,
 );
 
 
-sub as_hash {
-    my ($self) = @_;
+has country_name => (
+    is => 'ro',
+    isa => Str,
+    required => 0,
+);
 
-    return {
-        CountryCode => (
-            $self->country_code,
-        ),
-        ($self->postal_code ?
-             (Postalcode => (
-                 $self->postal_code,
-             )) : () ),
-        ($self->city ?
-             (City => (
-                 $self->city,
-             )) : () ),
+
+{
+our $_self;
+sub _if {
+    my ($method,$key) = @_;
+    if ($_self->$method) {
+        return ( $key => $_self->$method );
+    }
+    return;
+};
+
+sub as_hash {
+    my ($self,$shape) = @_;
+    local $_self=$self;
+
+    if ($shape eq 'capability') {
+        return {
+            _if(postal_code => 'Postalcode'),
+            _if(city => 'City'),
+            CountryCode => $self->country_code,
+        };
+    }
+    elsif ($shape eq 'route') {
+        return {
+            _if(line1 => 'Address1'),
+            _if(line2 => 'Address2'),
+            _if(line3 => 'Address3'),
+            _if(postal_code => 'PostalCode'),
+            _if(city => 'City'),
+            _if(division => 'Division'),
+            CountryCode => $self->country_code,
+            CountryName => '', # the value is required, but an empty
+                               # string will do
+            _if(country_name => 'CountryName'),
+        };
     };
 }
+}
 
 1;
 
@@ -64,14 +107,26 @@ Net::Async::Webservice::DHL::Address - an address for DHL
 
 =head1 VERSION
 
-version 0.01_6
+version 0.001
 
 =head1 ATTRIBUTES
 
+=head2 C<line1>
+
+=head2 C<line2>
+
+=head2 C<line3>
+
+Address lines, all optional strings.
+
 =head2 C<city>
 
 String with the name of the city, optional.
 
+=head2 C<division>
+
+Code of the division (e.g. state, prefecture, etc.), optional string.
+
 =head2 C<postal_code>
 
 String with the post code of the address, optional.
@@ -80,6 +135,10 @@ String with the post code of the address, optional.
 
 String with the 2 letter country code, required.
 
+=head2 C<country_name>
+
+String with the full country name, required only for some uses.
+
 =head1 METHODS
 
 =head2 C<as_hash>
